import deap.tools
import deap.algorithms

def elitist_ea(population, toolbox, nelite, cxpb, mutpb, ngen, stats=None,
               halloffame=None, verbose=__debug__):
    """This algorithm extends the eaSimple of deap (https://github.com/DEAP/deap/blob/master/deap/algorithms.py).
    See that algorithm for more documentation.
    This algorithm also introduces a new parameter nelite
    :param nelite: The number of elite individuals that are carried into the next generation without any modification.
    """
    logbook = deap.tools.Logbook()
    logbook.header = ['gen', 'nevals'] + (stats.fields if stats else [])

    # Evaluate the individuals with an invalid fitness
    invalid_ind = [ind for ind in population if not ind.fitness.valid]
    fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
    for ind, fit in zip(invalid_ind, fitnesses):
        ind.fitness.values = fit

    if halloffame is not None:
        halloffame.update(population)

    record = stats.compile(population) if stats else {}
    logbook.record(gen=0, nevals=len(invalid_ind), **record)
    if verbose:
        print(logbook.stream)

    # Begin the generational process
    for gen in range(1, ngen + 1):
        # Create the elite set
        sorted_population = sorted(population, key=lambda x: x.fitness)
        elite = sorted_population[0:nelite]

        # Select the next generation individuals
        offspring = toolbox.select(population, len(population)-nelite)

        # Vary the pool of individuals
        offspring = deap.algorithms.varAnd(offspring, toolbox, cxpb, mutpb)

        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit

        # Update the hall of fame with the generated individuals
        if halloffame is not None:
            halloffame.update(offspring)

        # Replace the current population by the offspring
        population[:] = elite + offspring

        # Append the current generation statistics to the logbook
        record = stats.compile(population) if stats else {}
        logbook.record(gen=gen, nevals=len(invalid_ind), **record)
        if verbose:
            print(logbook.stream)

    return population, logbook
