import random

from deap import base
from deap import creator
from deap import tools
from deap import gp

import numpy

import behavior_tree as bt

import genetic_programming.jones


# primitives for the inner nodes
def sequencestar1child(selector):
    return bt.SequenceStarNode([selector])


def sequencestar2children(selector1, selector2):
    return bt.SequenceStarNode([selector1, selector2])


def sequencestar3children(selector1, selector2, selector3):
    return bt.SequenceStarNode([selector1, selector2, selector3])


def sequencestar4children(selector1, selector2, selector3, selector4):
    return bt.SequenceStarNode([selector1, selector2, selector3, selector4])


def selector2children(condition, action):
    return bt.SelectorNode(condition, action)


# primitives for the leaf nodes
def condition_black_floor(beta):
    params = {"p": beta}
    return bt.ConditionNode(0, params)


def condition_gray_floor(beta):
    params = {"p": beta}
    return bt.ConditionNode(1, params)


def condition_white_floor(beta):
    params = {"p": beta}
    return bt.ConditionNode(2, params)


def condition_fixed_probability(beta):
    params = {"p": beta}
    return bt.ConditionNode(5, params)


def condition_neighborcount(eta, xi):
    params = {"w": eta, "p": xi}
    return bt.ConditionNode(3, params)


def condition_inverted_neighborcount(eta, xi):
    params = {"w": eta, "p": xi}
    return bt.ConditionNode(4, params)


class ParamBeta(float):
    pass


class ParamXi(int):
    pass


class ParamEta(float):
    pass


def action_exploration(rwm):
    params = {"rwm": rwm}
    return bt.ActionNode(0, params)


def action_stop():
    params = {}
    return bt.ActionNode(1, params)


def action_phototaxis():
    params = {}
    return bt.ActionNode(2, params)


def action_antiphototaxis():
    params = {}
    return bt.ActionNode(3, params)


def action_attraction(att):
    params = {"att": att}
    return bt.ActionNode(4, params)


def action_repulsion(rep):
    params = {"rep": rep}
    return bt.ActionNode(5, params)


class ParamAttRep(float):
    pass


class ParamRWM(int):
    pass


def create_primitive_set():
    # defined a new primitive set for strongly typed GP
    pset = gp.PrimitiveSetTyped("MAIN", [], bt.SequenceStarNode)

    # operator for sequence* node
    pset.addPrimitive(sequencestar1child, [bt.SelectorNode], bt.SequenceStarNode)
    pset.addPrimitive(sequencestar2children, [bt.SelectorNode, bt.SelectorNode], bt.SequenceStarNode)
    pset.addPrimitive(sequencestar3children, [bt.SelectorNode, bt.SelectorNode, bt.SelectorNode], bt.SequenceStarNode)
    pset.addPrimitive(sequencestar4children, [bt.SelectorNode, bt.SelectorNode, bt.SelectorNode, bt.SelectorNode], bt.SequenceStarNode)

    # operator for selector node
    # pset.addPrimitive(selector0children, [], SelectorNode)
    pset.addPrimitive(selector2children, [bt.ConditionNode, bt.ActionNode], bt.SelectorNode)

    # operator for condition node
    pset.addPrimitive(condition_black_floor, [ParamBeta], bt.ConditionNode)
    pset.addPrimitive(condition_gray_floor, [ParamBeta], bt.ConditionNode)
    pset.addPrimitive(condition_white_floor, [ParamBeta], bt.ConditionNode)
    pset.addPrimitive(condition_neighborcount, [ParamEta, ParamXi], bt.ConditionNode)
    pset.addPrimitive(condition_inverted_neighborcount, [ParamEta, ParamXi], bt.ConditionNode)
    pset.addPrimitive(condition_fixed_probability, [ParamBeta], bt.ConditionNode)

    # operator for action node
    pset.addPrimitive(action_exploration, [ParamRWM], bt.ActionNode)
    pset.addPrimitive(action_stop, [], bt.ActionNode)
    pset.addPrimitive(action_phototaxis, [], bt.ActionNode)
    pset.addPrimitive(action_antiphototaxis, [], bt.ActionNode)
    pset.addPrimitive(action_attraction, [ParamAttRep], bt.ActionNode)
    pset.addPrimitive(action_repulsion, [ParamAttRep], bt.ActionNode)

    # terminals
    pset.addEphemeralConstant("param_beta", lambda: random.uniform(0, 1), ParamBeta)
    pset.addEphemeralConstant("param_eta", lambda: random.uniform(0, 20), ParamEta)
    pset.addEphemeralConstant("param_xi", lambda: random.randint(0, 10), ParamXi)

    pset.addEphemeralConstant("param_rwm", lambda: random.randint(0, 100), ParamRWM)
    pset.addEphemeralConstant("param_att_rep", lambda: random.uniform(1, 5), ParamAttRep)

    return pset


def create_toolbox(pset):
    toolbox = base.Toolbox()
    toolbox.register("expr", gp.genHalfAndHalf, pset=pset, min_=3, max_=3)
    toolbox.register("individual", tools.initIterate, creator.Individual, toolbox.expr)
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    toolbox.register("compile", gp.compile, pset=pset)

    toolbox.register("evaluate", evalBT, toolbox=toolbox)

    return toolbox


def create_hall_of_fame():
    hof = tools.HallOfFame(1)
    return hof


def create_stats():
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("avg", numpy.mean)
    stats.register("std", numpy.std)
    stats.register("min", numpy.min)
    stats.register("max", numpy.max)
    return stats


def setup():
    creator.create("FitnessMax", base.Fitness, weights=(1.0,))
    creator.create("Individual", gp.PrimitiveTree, fitness=creator.FitnessMax)

    pset = create_primitive_set()
    toolbox = create_toolbox(pset)
    hof = create_hall_of_fame()
    stats = create_stats()

    return pset, toolbox, hof, stats


def compute_subtrees(individual, index=0):
    """
    Returns the nodes defining the subtrees of the node defined by index. Default is the top-level node
    :param individual:
    :param index:
    :return:
    """
    current_pos = index
    subtrees = []
    for _ in range(individual[index].arity):
        subtree_slice = individual.searchSubtree(current_pos + 1)
        subtree = individual[subtree_slice]  # get the ith subtree
        subtrees.append(subtree)
        current_pos += len(subtree)
    return subtrees


def compute_subtree_indices(individual, index=0):
    """
    Computes the direct subtrees of the node defined by index. Default is the top-level node
    :param individual:
    :param index:
    :return:
    """
    current_pos = index
    subtree_indices = []
    for _ in range(individual[index].arity):
        subtree_slice = individual.searchSubtree(current_pos + 1)
        subtree = individual[subtree_slice]  # get the ith subtree
        subtree_indices.append(current_pos+1)
        current_pos += len(subtree)
    return subtree_indices


def remove_subtree(individual, subtree_index):
    """
    Removes the subtree defined by the node at subtree_index
    :param individual:
    :param subtree_index:
    :return:
    """
    subtree_slice = individual.searchSubtree(subtree_index)
    for _ in range(subtree_slice.start, subtree_slice.stop):
        individual.pop(subtree_index)


def print_individual(individual):
    print(individual)
    for node in individual:
        print("    {} ({}): {}".format(node.name, node.arity, node.args))


def evalBT(individual, toolbox=None):
    # Transform the tree expression in a callable function
    toplevel = toolbox.compile(expr=individual)
    # toplevel = func()
    scores = bt.evaluate_controller(toplevel)
    return numpy.mean(scores),


budget = 0
