import random
import copy

from deap import gp
from deap import tools

import custom_algorithms
import genetic_programming
import behavior_tree


def restricted_mutUniform(individual, pset):
    """
    Extending the mutUniform of deap.gp
    :param individual:
    :param pset:
    """
    # get a random subtree
    index = random.randrange(len(individual))
    slice_ = individual.searchSubtree(index)
    # get its return type
    type_ = individual[index].ret
    # get its depth
    stack = []
    for node in individual[0:index]:
        # TODO: implement calculating the depth with a stack
        pass
    # TODO: remove this hacky way of estimating the remaining depth:
    if type_ == behavior_tree.SequenceStarNode:
        remaining_depth = 3
    elif type_ == behavior_tree.SelectorNode:
        remaining_depth = 2
    elif type_ == behavior_tree.ActionNode or type_ == behavior_tree.ConditionNode:
        remaining_depth = 1
    else:
        remaining_depth = 0
    individual[slice_] = gp.genFull(pset=pset, type_=type_, min_=remaining_depth, max_=remaining_depth)
    return


def restricted_mutShrink(individual, index=0):
    """
    Shrink the current tree by removing the selector-subtree defined by index.
    If index=0 then a random selector subtree is removed.
    Contains some sanity checks.
    :param individual:
    :param index:
    :return:
    """
    subtree_indices = genetic_programming.compute_subtree_indices(individual)
    if len(subtree_indices) == 1:
        print("Length of subtrees was equal to 1; can't shrink")
        return
    if index > 0 and index not in subtree_indices:
        print("Index {} is not the index of a selector-subtree".format(index))
        return
    # remove a random subtree
    shrink_tree_index = random.choice(subtree_indices)
    subtree_indices.remove(shrink_tree_index)
    genetic_programming.remove_subtree(individual, shrink_tree_index)
    # update the top-level node
    top_level = copy.copy(individual[0])  # make a copy so that changes to break anything in the other individuals
    individual[0] = top_level  # the copy needs to be assigned before changes can be made
    top_level.name = "sequencestar{}child{}".format(len(subtree_indices), "ren" if len(subtree_indices) > 1 else "")
    top_level.args = [top_level.args[0]] * len(subtree_indices)
    top_level.arity = len(subtree_indices)
    args = ", ".join(map("{{{0}}}".format, range(top_level.arity)))
    top_level.seq = "{name}({args})".format(name=top_level.name, args=args)
    # top_level.ret stays the same
    return


def restricted_mutNodeReplacement(individual, pset):
    """
    This method selects and action or condition node at random and replaces it with another condition or action node.
    :param individual: The individual to be mutated
    :param pset: A typed primitive set, used to choose the the replacement node
    """
    # select a random action or condition node
    action_condition_list = [x for x in enumerate(individual)
                             if x[1].ret == behavior_tree.ActionNode or x[1].ret == behavior_tree.ConditionNode]
    replace_node = random.choice(action_condition_list)
    return_type = replace_node[1].ret
    slice_ = individual.searchSubtree(replace_node[0])
    individual[slice_] = gp.genFull(pset=pset, type_=return_type, min_=1, max_=1)
    return


def jones_restricted_mutation(individual, toolbox=None):
    prob_mut_u = 0.05  # uniform
    prob_mut_s = 0.1  # shrink
    prob_mut_n = 0.5  # node replacement
    prob_mut_e = 0.5  # ephemeral constant

    # print(individual[0].name)  # name is a string
    # print(individual[0].args)  # args is a list of classes (the classes for the arguments)
    # print(individual[0].arity)  # arity is an int
    # print(individual[0].ret)  # ret is the return type

    # mutUniform is transformed into the following:
    # choose a random node in the tree
    # if ephemeral constant: mutate the constant
    # if action or condition node: transform into different action/condition node
    # if selector_subtree: mutate both action and condition
    # if sequencestar node: create completely random tree
    if random.random() <= prob_mut_u:
        restricted_mutUniform(individual, toolbox.pset())

    # mutShrink is transformed into randomly removing one "selector-subtree"
    if random.random() <= prob_mut_s:
        restricted_mutShrink(individual)

    # mutNodeReplacement will select a random leaf node and replace the associated action or behavior
    if random.random() <= prob_mut_n:
        restricted_mutNodeReplacement(individual, toolbox.pset())

    # mutEphemeral is kept
    if random.random() <= prob_mut_e:
        gp.mutEphemeral(individual, "one")

    return individual,


def gp_jones():
    pset, toolbox, hof, stats = genetic_programming.setup()
    toolbox.register("select", tools.selTournament, tournsize=3)  # 1% of population size
    toolbox.register("mate", gp.cxOnePoint)
    toolbox.register("mutate", jones_restricted_mutation, toolbox=toolbox)
    toolbox.register("mutUniformExpression", restricted_mutUniform)
    toolbox.register("pset", lambda: pset)

    pop_size = 25
    number_of_elites = 3

    pop = toolbox.population(n=pop_size)

    prob_crossover = 0.8
    prob_mutation = 1  # always call the custom mutation function, the real mutation probabilities are set there
    generations = genetic_programming.budget // (10 * pop_size)
    print("Running for {} generations".format(generations))

    custom_algorithms.elitist_ea(pop, toolbox, number_of_elites, prob_crossover, prob_mutation, generations, stats, halloffame=hof)

    best_individual = toolbox.compile(expr=hof[0]).to_cmd()

    print("Best controller found (Hall of Fame):")
    print(best_individual)

    return pop, stats, hof
