#!/bin/bash
###############################################################################
# This script launches several runs of irace in parallel in a SGE Cluster, and
# each run is parallelized using MPI.  Execute without parameters to see usage.
###############################################################################
set -e
set -o pipefail

#QUEUE=short
QUEUE=long

#MACHINE=opteron2216 #rack 1
#MACHINE=xeon5410 #rack 2
#MACHINE=opteron6128 #rack 3
#MACHINE=opteron6272 #rack 4
MACHINE=xeon2680 #rack 5
#MACHINE=xeon6138 #rack 6

# This function launches one run of irace.
launch_bt_gp() {
    JOBNAME=BT-GP-$$-$1
    CONFIG=$2
    SCENARIO=$3
    exec qsub -v PATH <<EOF
#!/bin/bash
#$ -N $JOBNAME
#$ -l $MACHINE
#$ -l $QUEUE
#$ -m ase
#      b     Mail is sent at the beginning of the job.
#      e     Mail is sent at the end of the job.
#      a     Mail is sent when the job is aborted or rescheduled.
#      s     Mail is sent when the job is suspended.
#
#$ -o $EXECDIR/BT-GP.stdout 
#$ -e $EXECDIR/BT-GP.stderr 
#$ -cwd
USERNAME=`whoami`
TMPDIR=/tmp/\${USERNAME}/BT-GP/$EXECDIR
JOBDIR=/home/\${USERNAME}/BehaviorTree-GeneticProgramming/$EXECDIR

mkdir -p \${TMPDIR}
source venv/bin/activate
python3 bt_gp.py -c ${CONFIG} -s ${SCENARIO} -b $4 2>&1 | tee $EXECDIR/BT-GP.out
RET=\$?
mv \${TMPDIR}/* \${JOBDIR}
cd \${JOBDIR}
rmdir -p \${TMPDIR} &> /dev/null
exit \${RET}
EOF
}
## End of customization

error () {
    echo "$0: error: $@" >&2
    exit 1
}

usage() {
    cat <<EOF
Usage: $0 N[-M] EXECDIR

Parameters:
 N                 an integer giving the number of repetitions of the genetic programming
                   or a sequence N-M giving which repetitions to redo.
 EXECDIR           job M will use EXECDIR-M directory
EOF
    exit 1
}

# Issue usage if no parameters are given.
test $# -ge 1 || usage

# Number of repetitions of iRace
REPETITIONS=$1
shift
START=1

if [[ "$REPETITIONS" =~ ^([0-9]+)-([0-9]+)$ ]] ; then
    START=${BASH_REMATCH[1]}
    REPETITIONS=${BASH_REMATCH[2]}
elif ! [[ "$REPETITIONS" =~ ^[0-9]+$ ]] ; then
    error "number of repetitions must be an integer"
fi

# execDir (--exec-dir) directory
EXECDIR_PREFIX=${1:-execdir}
shift

BUDGET=${1:-budget}
shift

for i in $(seq $START $REPETITIONS); do
    EXECDIR=$(printf '%s-%002d-for-%00005d' ${EXECDIR_PREFIX} $i $BUDGET)
    echo "execution directory: $EXECDIR"
    rm -rf $EXECDIR
    mkdir -p $EXECDIR
    launch_bt_gp $(printf '%002d' $i) $EXECDIR_PREFIX "/home/jkuckling/AutoMoDe-missions/vanilla/for120sec/foraging_bt.argos" $BUDGET &
    sleep 1

    EXECDIR=$(printf '%s-%002d-agg-%00005d' ${EXECDIR_PREFIX} $i $BUDGET)
    echo "execution directory: $EXECDIR"
    rm -rf $EXECDIR
    mkdir -p $EXECDIR
    launch_bt_gp $(printf '%002d' $i) $EXECDIR_PREFIX "/home/jkuckling/AutoMoDe-missions/vanilla/agg250sec/aggregation_bt.argos" $BUDGET &
    sleep 1

    EXECDIR=$(printf '%s-%002d-aac-%00005d' ${EXECDIR_PREFIX} $i $BUDGET)
    echo "execution directory: $EXECDIR"
    rm -rf $EXECDIR
    mkdir -p $EXECDIR
    launch_bt_gp $(printf '%002d' $i) $EXECDIR_PREFIX "/home/jkuckling/AutoMoDe-missions/chocolate/AAC120sec/aac_bt.argos" $BUDGET &
    sleep 1

    EXECDIR=$(printf '%s-%002d-sca-%00005d' ${EXECDIR_PREFIX} $i $BUDGET)
    echo "execution directory: $EXECDIR"
    rm -rf $EXECDIR
    mkdir -p $EXECDIR
    launch_bt_gp $(printf '%002d' $i) $EXECDIR_PREFIX "/home/jkuckling/AutoMoDe-missions/chocolate/SCA120sec/sca_bt.argos" $BUDGET &
    sleep 1

    EXECDIR=$(printf '%s-%002d-guided_shelter-%00005d' ${EXECDIR_PREFIX} $i $BUDGET)
    echo "execution directory: $EXECDIR"
    rm -rf $EXECDIR
    mkdir -p $EXECDIR
    launch_bt_gp $(printf '%002d' $i) $EXECDIR_PREFIX "/home/jkuckling/AutoMoDe-missions/guided_shelter120sec/guided_shelter_bt.argos" $BUDGET &
    sleep 1
done




