"""
This module tests the implementation of the mutation operators for the approach after Jones et al.
"""

import unittest

import genetic_programming
import genetic_programming.jones


class TestJonesMutationCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """
        Handles the setup for the gp framework.
        Creates the primitive set (pset), the toolbox, the hall of fame (hof) and the statistics object (stats)
        """
        cls.pset, cls.toolbox, cls.hof, cls.stats = genetic_programming.setup()


class TestJonesMutationShrink(TestJonesMutationCase):
    """
    This class contains all test methods for the restricted_mutShrink operator
    """

    def setUp(self):
        """
        Handles the setup for each test case.
        Creates an individual with more than one selector-subtree (otherwise it can't be shrunken).
        """
        # create an individual that has more than one subtree
        self.individual = self.toolbox.individual()
        while self.individual[0].name == "sequencestar1child":
            self.individual = self.toolbox.individual()

    def test_tree_size(self):
        """
        Simple test to assert that self.individual has indeed more than one subtree
        """
        self.assertGreater(len(genetic_programming.compute_subtrees(self.individual)), 1)

    def shrink(self):
        """
        Since all the methods here rely on the same method for shrinking, I have moved it here. It allows to modify the
        shrinking simultaneously for all test.
        """
        genetic_programming.jones.restricted_mutShrink(self.individual, 1)

    def test_new_length(self):
        """
        This test asserts that the mutated individual has exactly one less subtree than before
        """
        old_subtree_length = len(genetic_programming.compute_subtrees(self.individual))
        self.shrink()
        new_subtree_length = len(genetic_programming.compute_subtrees(self.individual))
        self.assertEqual(old_subtree_length, new_subtree_length + 1)

    def test_arity(self):
        """
        This test asserts that the mutated individual's first element's arity is the same as the number of subtrees
        """
        self.shrink()
        self.assertEqual(self.individual[0].arity, len(genetic_programming.compute_subtrees(self.individual)))

    def test_isolation(self):
        """
        This test asserts that a copy of the individual is not modified if the individual is mutated.
        It could be extended to cover more properties but the arity is already a good indicator.
        """
        copied_individual = TestJonesMutationShrink.toolbox.clone(self.individual)
        self.shrink()
        self.assertNotEqual(self.individual[0].arity, copied_individual[0].arity)
